"""ioHub Common Eye Tracker Interface for EyeLink(C) Systems"""
 # Part of the psychopy.iohub library.
 # Copyright (C) 2012-2016 iSolver Software Solutions
 # Distributed under the terms of the GNU General Public License (GPL).