#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file was generated by generateHints.py.
# Following strings are used to localize hints in Preference Dialog of
# the PsychoPy application.
# Rebuild this file if comments in *.spec files are modified.

from __future__ import absolute_import, print_function
from psychopy.localization import _translate

# baseNoArch.spec,[general],line25
_translate("which system to use as a backend for drawing")

# baseNoArch.spec,[general],line27
_translate("the default units for windows and visual stimuli")

# baseNoArch.spec,[general],line29
_translate("full screen is best for accurate timing")

# baseNoArch.spec,[general],line31
_translate("enable subjects to use the mouse and GUIs during experiments")

# baseNoArch.spec,[general],line33
_translate("'version' is for internal usage, not for the user")

# baseNoArch.spec,[general],line35
_translate("Add paths here to your custom Python modules")

# baseNoArch.spec,[general],line37
_translate("path to flac (lossless audio compression) on this operating system")

# baseNoArch.spec,[general],line39
_translate("Shutdown keys, following the pyglet naming scheme.")

# baseNoArch.spec,[general],line41
_translate("Modifier keys for shutdown keys")

# baseNoArch.spec,[general],line43
_translate("What to do if gamma-correction not possible")

# baseNoArch.spec,[general],line45
_translate("Add plugin names here to load when a PsychoPy session starts.")

# baseNoArch.spec,[general],line48
_translate("display tips when starting PsychoPy")

# baseNoArch.spec,[app],line50
_translate("what windows to display when PsychoPy starts")

# baseNoArch.spec,[app],line52
_translate("reset preferences to defaults on next restart of PsychoPy")

# baseNoArch.spec,[app],line54
_translate("save any unsaved preferences before closing the window")

# baseNoArch.spec,[app],line56
_translate("enable features for debugging PsychoPy itself, including unit-tests")

# baseNoArch.spec,[app],line58
_translate("language to use in menus etc; not all translations are available. Select a value, then restart the app.")

# baseNoArch.spec,[app],line60
_translate(
    "Show an error dialog when PsychoPy encounters an unhandled internal error.")

# baseNoArch.spec,[app],line62
_translate("Theme")

# baseNoArch.spec,[app],line64
_translate(
    "open Coder files as read-only (allows running without accidental changes)")

# baseNoArch.spec,[app],line67
_translate("a list of font names; the first one found on the system will be used")

# baseNoArch.spec,[coder],line69
_translate("Font size (in pts) takes an integer between 6 and 24")

# baseNoArch.spec,[coder],line71
_translate("Spacing between lines")

# baseNoArch.spec,[coder],line75
_translate("Long line edge guide, specify zero to disable")

# baseNoArch.spec,[coder],line79
_translate("Set the source assistant panel to be visible by default")

# baseNoArch.spec,[coder],line81
_translate("Set the output/shell to be visible by default")

# baseNoArch.spec,[coder],line83
_translate(
    "Show code completion suggestion and calltips automatically when typing.")

# baseNoArch.spec,[coder],line85
_translate("reload previously opened files after start")

# baseNoArch.spec,[coder],line87
_translate("for coder shell window, which shell to use")

# baseNoArch.spec,[coder],line89
_translate("whether to automatically reload a previously open experiment")

# baseNoArch.spec,[coder],line91
_translate("Default to when writing code components")

# baseNoArch.spec,[coder],line94
_translate(
    "if False will create scripts with an 'easier' but more cluttered namespace")

# baseNoArch.spec,[builder],line96
_translate("folder names for custom components; expects a comma-separated list")

# baseNoArch.spec,[builder],line98
_translate("a list of components to hide (eg, because you never use them)")

# baseNoArch.spec,[builder],line100
_translate("where the Builder demos are located on this computer (after unpacking)")

# baseNoArch.spec,[builder],line102
_translate(
    "name of the folder where subject data should be saved (relative to the script)")

# baseNoArch.spec,[builder],line104
_translate("Panels arrangement: Should Flow be on the top or bottom, and should Components be on the left or right?")

# baseNoArch.spec,[builder],line106
_translate("Display text in a floating window that describes the experiment")

# baseNoArch.spec,[builder],line108
_translate("choice of audio library")

# baseNoArch.spec,[builder],line110
_translate(
    "latency mode for PsychToolbox audio (3 is good for most applications. See")

# baseNoArch.spec,[builder],line112
_translate("audio driver to use")

# baseNoArch.spec,[builder],line114
_translate("audio device to use (if audioLib allows control)")

# baseNoArch.spec,[hardware],line118
_translate("a list of parallel ports")

# baseNoArch.spec,[hardware],line120
_translate("The name of the Qmix pump configuration to use")

# baseNoArch.spec,[hardware],line122
_translate("the http proxy for usage stats and auto-updating; format is host: port")

# baseNoArch.spec,[hardware],line124
_translate(
    "override the above proxy settings with values found in the environment (if possible)")

# baseNoArch.spec,[hardware],line126
_translate("allow PsychoPy to send anonymous usage stats; please allow if possible, it helps PsychoPy's development")

# baseNoArch.spec,[hardware],line128
_translate("allow PsychoPy to check for new features and bug fixes")

# baseNoArch.spec,[hardware],line131
_translate("max time to wait for a connection response")

# baseNoArch.spec,[connections],line133
_translate("open an existing file")

# baseNoArch.spec,[connections],line135
_translate("start a new experiment or script")

# baseNoArch.spec,[connections],line137
_translate("save a Builder or Coder file")

# baseNoArch.spec,[connections],line139
_translate("save a Builder or Coder file under a new name")

# baseNoArch.spec,[connections],line141
_translate("Coder: print the file")

# baseNoArch.spec,[connections],line144
_translate("close the Builder or Coder window")

# baseNoArch.spec,[keyBindings],line146
_translate("end the application (PsychoPy)")

# baseNoArch.spec,[keyBindings],line148
_translate("open the preferences dialog")

# baseNoArch.spec,[keyBindings],line150
_translate("export Builder experiment to HTML")

# baseNoArch.spec,[keyBindings],line152
_translate("Coder: cut")

# baseNoArch.spec,[keyBindings],line154
_translate("Coder: copy")

# baseNoArch.spec,[keyBindings],line156
_translate("Coder: paste")

# baseNoArch.spec,[keyBindings],line158
_translate("Coder: duplicate")

# baseNoArch.spec,[keyBindings],line160
_translate("Coder: indent code by one level (4 spaces)")

# baseNoArch.spec,[keyBindings],line162
_translate("Coder: reduce indentation by one level (4 spaces)")

# baseNoArch.spec,[keyBindings],line165
_translate("Coder: indent to fit python syntax")

# baseNoArch.spec,[keyBindings],line167
_translate("Coder: find")

# baseNoArch.spec,[keyBindings],line169
_translate("Coder: find again")

# baseNoArch.spec,[keyBindings],line171
_translate("Coder: undo")

# baseNoArch.spec,[keyBindings],line173
_translate("Coder: redo")

# baseNoArch.spec,[keyBindings],line175
_translate("Coder: add a # to the start of the line(s)")

# baseNoArch.spec,[keyBindings],line177
_translate("Coder: remove # from start of line(s)")

# baseNoArch.spec,[keyBindings],line179
_translate("Coder: add or remove # from start of line(s)")

# baseNoArch.spec,[keyBindings],line181
_translate("Coder: fold this block of code")

# baseNoArch.spec,[keyBindings],line183
_translate("Coder: increase font size this block of code")

# baseNoArch.spec,[keyBindings],line185
_translate("Coder: decrease font size this block of code")

# baseNoArch.spec,[keyBindings],line187
_translate("Coder: check for basic syntax errors")

# baseNoArch.spec,[keyBindings],line189
_translate(
    "convert a Builder .psyexp script into a python script and open it in the Coder")

# baseNoArch.spec,[keyBindings],line191
_translate("launch a script, Builder or Coder, or run unit-tests")

# baseNoArch.spec,[keyBindings],line193
_translate("attempt to interrupt and halt a running script")

# baseNoArch.spec,[keyBindings],line195
_translate("Coder: show / hide white-space dots")

# baseNoArch.spec,[keyBindings],line197
_translate("Coder: show / hide end of line characters")

# baseNoArch.spec,[keyBindings],line200
_translate("Coder: show / hide indentation level lines")

# baseNoArch.spec,[keyBindings],line202
_translate("Builder: create a new routine")

# baseNoArch.spec,[keyBindings],line204
_translate("Builder: copy an existing routine")

# baseNoArch.spec,[keyBindings],line208
_translate("Builder: paste the copied routine")

# baseNoArch.spec,[keyBindings],line211
_translate("Builder: paste the copied component")

# baseNoArch.spec,[keyBindings],line213
_translate("Coder: show / hide the output panel")

# baseNoArch.spec,[keyBindings],line215
_translate("Builder: rename an existing routine")

# baseNoArch.spec,[keyBindings],line218
_translate("switch between windows")

# baseNoArch.spec,[keyBindings],line220
_translate("increase display size in Flow")

# baseNoArch.spec,[keyBindings],line222
_translate("decrease display size in Flow")

# baseNoArch.spec,[keyBindings],line224
_translate("increase display size of Routines")

# baseNoArch.spec,[keyBindings],line226
_translate("decrease display size of Routines")

# baseNoArch.spec,[keyBindings],line228
_translate("show or hide the readme (info) for this experiment if possible")

# baseNoArch.spec,[keyBindings],line230
_translate("Projects: Log in to pavlovia")

# baseNoArch.spec,[keyBindings],line232
_translate("Projects: Log in to OSF")

# baseNoArch.spec,[keyBindings],line234
_translate("Projects: Sync project")

# baseNoArch.spec,[keyBindings],line236
_translate("Projects: Find projects")

# baseNoArch.spec,[keyBindings],line238
_translate("Projects: Open project")

# baseNoArch.spec,[keyBindings],line240
_translate("Projects: Create new project")
