"""ioHub Common Eye Tracker Interface"""

# Part of the psychopy.iohub library.
# Copyright (C) 2012-2016 iSolver Software Solutions
# Distributed under the terms of the GNU General Public License (GPL).

from __future__ import absolute_import
from .eyetracker import *
