#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""Functions and classes related to unit conversion"""

# This module is not used by psychopy; retained for backwards compatibility
# for user-scripts.

# pylint: disable=W0611
# W0611 = Unused import %s

from __future__ import absolute_import, print_function

from numpy import radians, degrees
